@@ -1,4 +1,4 @@
-# Copyright 2018 Hewlett Packard Enterprise Development LP
+# Copyright 2018-2022 Hewlett Packard Enterprise Development LP
 #
 # Licensed under the Apache License, Version 2.0 (the "License"); you may
 # not use this file except in compliance with the License. You may obtain
@@ -14,9 +14,12 @@
 
 __author__ = 'HPE'
 
+from base64 import b64decode
 import json
 import re
 
+from OpenSSL.crypto import FILETYPE_ASN1
+from OpenSSL.crypto import load_certificate
 from six.moves.urllib import parse
 import sushy
 from sushy.resources.system import mappings as sushy_map
@@ -28,6 +31,7 @@ from proliantutils.ilo import firmware_controller
 from proliantutils.ilo import operations
 from proliantutils import log
 from proliantutils.redfish import main
+from proliantutils.redfish.resources import gpu_common
 from proliantutils.redfish.resources.manager import constants as mgr_cons
 from proliantutils.redfish.resources.system import constants as sys_cons
 from proliantutils.redfish.resources.system.storage \
@@ -98,6 +102,7 @@ GET_POST_STATE_MAP = {
 # collection, as we are dealing with iLO's here.
 PROLIANT_MANAGER_ID = '1'
 PROLIANT_SYSTEM_ID = '1'
+PROLIANT_CHASSIS_ID = '1'
 
 BOOT_OPTION_MAP = {'BOOT_ONCE': True,
                    'BOOT_ALWAYS': False,
@@ -114,6 +119,9 @@ SUPPORTED_BOOT_MODE_MAP = {
         ilo_cons.SUPPORTED_BOOT_MODE_LEGACY_BIOS_AND_UEFI)
 }
 
+_CERTIFICATE_PATTERN = (
+    r'-----BEGIN CERTIFICATE-----.*?-----END CERTIFICATE-----')
+
 LOG = log.get_logger(__name__)
 
 
@@ -208,6 +216,24 @@ class RedfishOperations(operations.IloOperations):
             LOG.debug(msg)
             raise exception.IloError(msg)
 
+    def _get_sushy_chassis(self, chassis_id):
+        """Get the sushy chassis for chassis_id
+
+        :param chassis_id: The identity of the Chassis resource
+        :returns: the Sushy Chassis instance
+        :raises: IloError
+        """
+        chassis_url = parse.urljoin(self._sushy.get_chassis_collection_path(),
+                                    chassis_id)
+        try:
+            return self._sushy.get_chassis(chassis_url)
+        except sushy.exceptions.SushyError as e:
+            msg = (self._('The Redfish Chassis "%(chassis)s" was not found. '
+                          'Error %(error)s') %
+                   {'chassis': chassis_id, 'error': str(e)})
+            LOG.debug(msg)
+            raise exception.IloError(msg)
+
     def get_product_name(self):
         """Gets the product name of the server.
 
@@ -418,6 +444,8 @@ class RedfishOperations(operations.IloOperations):
         :raises: IloError, on an error from iLO.
         :raises: IloInvalidInputError, if the device is not valid.
         """
+        # Validate url
+        common_utils.validate_href(url)
         self._validate_virtual_media(device)
         manager = self._get_sushy_manager(PROLIANT_MANAGER_ID)
         try:
@@ -654,15 +682,279 @@ class RedfishOperations(operations.IloOperations):
             LOG.debug(msg)
             raise exception.IloError(msg)
 
+    def _update_security_parameter(self, sec_param, ignore=False):
+        """Sets the ignore flag for the security parameter.
+
+        :param sec_param: Name of the security parameter.
+        :param ignore : True when security parameter needs to be ignored.
+               If passed False, security param will not be ignored.
+               If nothing passed default will be False.
+        """
+        sushy_manager = self._get_sushy_manager(PROLIANT_MANAGER_ID)
+        try:
+            security_params = (
+                sushy_manager.securityservice.securityparamscollectionuri)
+            param_members = security_params.get_members()
+            for param in param_members:
+                if sec_param in param.name:
+                    param.update_security_param_ignore_status(ignore)
+                    break
+            else:
+                msg = (self._('Specified parameter "%(param)s" is not '
+                              'a Security Dashboard Parameter.') %
+                       {'param': sec_param})
+                raise exception.IloInvalidInputError(msg)
+        except sushy.exceptions.SushyError as e:
+            msg = (self._("The Redfish controller is unable to update "
+                          "resource or its member. Error "
+                          "%(error)s)") % {'error': str(e)})
+            LOG.debug(msg)
+            raise exception.IloError(msg)
+
+    def update_password_complexity(self, enable=True, ignore=False):
+        """Update the Password_Complexity security param.
+
+        :param enable: A boolean param, True when Password_Complexity needs
+               to be enabled. If passed False, Password_Complexity security
+               param will be disabled. If nothing passed default will be True.
+        :param ignore : A boolean param, True when Password_Complexity needs
+               to be ignored. If passed False, Password_Complexity security
+               param will not be ignored. If nothing passed default will be
+               False.
+        :raises: IloError, on an error from iLO.
+        """
+        acc_service = self._sushy.get_account_service()
+        try:
+            self._update_security_parameter(sec_param="Password Complexity",
+                                            ignore=ignore)
+            acc_service.update_enforce_passwd_complexity(enable)
+        except sushy.exceptions.SushyError as e:
+            msg = (self._('The Redfish controller failed to update the '
+                          'security dashboard parameter '
+                          '``Password_Complexity``. '
+                          'Error %(error)s') % {'error': str(e)})
+            LOG.debug(msg)
+            raise exception.IloError(msg)
+
+    def update_require_login_for_ilo_rbsu(self, enable=True, ignore=False):
+        """Update the RequiredLoginForiLORBSU security param.
+
+        :param enable: A boolean param, True when RequiredLoginForiLORBSU
+               needs to be enabled. If passed False, RequiredLoginForiLORBSU
+               security param will be disabled. If nothing passed default
+               will be True.
+        :param ignore : A boolean param, True when RequiredLoginForiLORBSU
+               needs to be ignored. If passed False, RequiredLoginForiLORBSU
+               security param will not be ignored. If nothing passed default
+               will be False.
+        :raises: IloError, on an error from iLO.
+        """
+        sushy_manager = self._get_sushy_manager(PROLIANT_MANAGER_ID)
+        try:
+            self._update_security_parameter(sec_param="Require Login",
+                                            ignore=ignore)
+            sushy_manager.update_login_for_ilo_rbsu(enable)
+        except sushy.exceptions.SushyError as e:
+            msg = (self._('The Redfish controller failed to update the '
+                          'security dashboard parameter '
+                          '``RequiredLoginForiLORBSU``. '
+                          'Error %(error)s') % {'error': str(e)})
+            LOG.debug(msg)
+            raise exception.IloError(msg)
+
+    def update_require_host_authentication(self, enable=True, ignore=False):
+        """Update the RequireHostAuthentication security param.
+
+        :param enable: A boolean param, True when RequireHostAuthentication
+               needs to be enabled. If passed False, RequireHostAuthentication
+               security param will be disabled. If nothing passed
+               default will be True.
+        :param ignore : A boolean param, True when RequireHostAuthentication
+               needs to be ignored. If passed False, RequireHostAuthentication
+               security param will not be ignored. If nothing passed
+               default will be False.
+        :raises: IloError, on an error from iLO.
+        """
+        sushy_manager = self._get_sushy_manager(PROLIANT_MANAGER_ID)
+        try:
+            self._update_security_parameter(sec_param="Host Authentication",
+                                            ignore=ignore)
+            sushy_manager.update_host_authentication(enable)
+        except sushy.exceptions.SushyError as e:
+            msg = (self._('The Redfish controller failed to update the '
+                          'security dashboard paramater '
+                          '``RequireHostAuthentication``. '
+                          'Error %(error)s') % {'error': str(e)})
+            LOG.debug(msg)
+            raise exception.IloError(msg)
+
+    def update_minimum_password_length(self, passwd_length=None, ignore=False):
+        """Update the MinPasswordLength security param.
+
+        :param passwd_length: Minimum lenght of password used. If nothing
+               passed default will be None.
+        :param ignore : A boolean param, True when MinPasswordLength needs to
+               be ignored. If passed False, MinPasswordLength security param
+               will not be ignored. If nothing passed default will be False.
+        """
+        acc_service = self._sushy.get_account_service()
+        try:
+            self._update_security_parameter(sec_param="Minimum",
+                                            ignore=ignore)
+            acc_service.update_min_passwd_length(passwd_length)
+        except sushy.exceptions.SushyError as e:
+            msg = (self._('The Redfish controller failed to update the '
+                          'security dashboard paramater '
+                          '``MinPasswordLength``. '
+                          'Error %(error)s') % {'error': str(e)})
+            LOG.debug(msg)
+            raise exception.IloError(msg)
+
+    def update_ipmi_over_lan(self, enable=False, ignore=False):
+        """Update the IPMI/DCMI_Over_LAN security param.
+
+        :param enable: A boolean param, True when IPMI/DCMI_Over_LAN needs to
+               be enabled. If passed False, IPMI/DCMI_Over_LAN security param
+               will be disabled. If nothing passed default will be False.
+        :param ignore : A boolean param, True when IPMI/DCMI_Over_LAN needs to
+               be ignored. If passed False, IPMI/DCMI_Over_LAN security param
+               will not be ignored. If nothing passed default will be False.
+        :raises: IloError, on an error from iLO.
+        """
+        sushy_manager = self._get_sushy_manager(PROLIANT_MANAGER_ID)
+        try:
+            self._update_security_parameter(sec_param="IPMI", ignore=ignore)
+            sushy_manager.networkprotocol.update_ipmi_enabled(enable)
+        except sushy.exceptions.SushyError as e:
+            msg = (self._('The Redfish controller failed to update the '
+                          'security dashboard paramater '
+                          '``IPMI/DCMI_Over_LAN``. '
+                          'Error %(error)s') % {'error': str(e)})
+            LOG.debug(msg)
+            raise exception.IloError(msg)
+
+    def update_authentication_failure_logging(self, logging_threshold=None,
+                                              ignore=False):
+        """Update the Authentication_failure_Logging security param.
+
+        :param logging_threshold: Value of authenication failure logging
+               threshold. If nothing passed default will be None.
+        :param ignore : A boolean param, True when
+               Authentication_failure_Logging needs to be ignored. If passed
+               False, Authentication_failure_Logging security param will not
+               be ignored. If nothing passed default will be False.
+        :raises: IloError, on an error from iLO.
+        """
+        acc_service = self._sushy.get_account_service()
+        try:
+            self._update_security_parameter(sec_param="Failure Logging",
+                                            ignore=ignore)
+            acc_service.update_auth_failure_logging(logging_threshold)
+        except sushy.exceptions.SushyError as e:
+            msg = (self._('The Redfish controller failed to update the '
+                          'security dashboard paramater '
+                          '``Authentication_failure_Logging``. '
+                          'Error %(error)s') % {'error': str(e)})
+            LOG.debug(msg)
+            raise exception.IloError(msg)
+
+    def update_secure_boot(self, enable=True, ignore=False):
+        """Update Secure_Boot security param on the server.
+
+        :param enable: A boolean param, True when Secure_Boot needs to be
+               enabled. If passed False, Secure_Boot security param will
+               be disabled. If nothing passed default will be True.
+        :param ignore : A boolean param, True when Secure_boot needs to be
+               ignored. If passed False, Secure_boot security param will
+               not be ignored. If nothing passed default will be False.
+        """
+        try:
+            self._update_security_parameter(sec_param="Secure Boot",
+                                            ignore=ignore)
+            self.set_secure_boot_mode(enable)
+        except sushy.exceptions.SushyError as e:
+            msg = (self._('The Redfish controller failed to update the '
+                          'security dashboard paramater ``Secure_boot``. '
+                          'Error %(error)s') % {'error': str(e)})
+            LOG.debug(msg)
+            raise exception.IloError(msg)
+
+    def get_security_dashboard_values(self):
+        """Gets all the parameters related to security dashboard.
+
+        :return: a dictionary of the security dashboard values
+                 with their security status and security parameters
+                 with their complete details and security status.
+        :raises: IloError, if security dashboard or their params
+            not found or on an error from iLO.
+        """
+        sec_capabilities = {}
+        sushy_manager = self._get_sushy_manager(PROLIANT_MANAGER_ID)
+        try:
+            security_dashboard = (
+                sushy_manager.securityservice.securitydashboard)
+            security_params = (
+                security_dashboard.securityparamscollectionuri)
+            if security_dashboard.server_configuration_lock_status:
+                sec_capabilities.update(
+                    {'server_configuration_lock_status': (
+                     security_dashboard.server_configuration_lock_status)})
+            sec_capabilities.update(
+                {'overall_security_status': (
+                 security_dashboard.overall_status)})
+            security_parameters = {}
+            param_members = security_params.get_members()
+            for param in param_members:
+                param_dict = {param.name: {'security_status': param.status,
+                                           'state': param.state,
+                                           'ignore': param.ignore}}
+                if param.description:
+                    param_dict[param.name].update(
+                        {'description': param.description})
+                if param.recommended_action:
+                    param_dict[param.name].update(
+                        {'recommended_action': param.recommended_action})
+                security_parameters.update(param_dict)
+            sec_capabilities.update(
+                {'security_parameters': security_parameters})
+        except sushy.exceptions.SushyError as e:
+            msg = (self._("The Redfish controller is unable to get "
+                          "resource or its members. Error "
+                          "%(error)s)") % {'error': str(e)})
+            LOG.debug(msg)
+            raise exception.IloError(msg)
+
+        return sec_capabilities
+
+    def _parse_security_dashboard_values_for_capabilities(self):
+        """Parses the security dashboard parameters.
+
+        :returns: a dictionary of only those security parameters and their
+            security status which are applicable for ironic.
+        """
+        values = self.get_security_dashboard_values()
+        ironic_sec_capabilities = {}
+        ironic_sec_capabilities.update(
+            {'overall_security_status': values.get('overall_security_status')})
+        param_values = values.get('security_parameters')
+        p_map = {'Last Firmware Scan Result': 'last_firmware_scan_result',
+                 'Security Override Switch': 'security_override_switch'}
+        p_keys = p_map.keys()
+        for p_key, p_val in param_values.items():
+            if p_key in p_keys:
+                p_dict = {p_map.get(p_key): p_val.get('security_status')}
+                ironic_sec_capabilities.update(p_dict)
+        return ironic_sec_capabilities
+
     def get_server_capabilities(self):
         """Returns the server capabilities
 
         raises: IloError on an error from iLO.
         """
         capabilities = {}
-
         sushy_system = self._get_sushy_system(PROLIANT_SYSTEM_ID)
         sushy_manager = self._get_sushy_manager(PROLIANT_MANAGER_ID)
+        sushy_chassis = self._get_sushy_chassis(PROLIANT_CHASSIS_ID)
         try:
             count = len(sushy_system.pci_devices.gpu_devices)
             boot_mode = rf_utils.get_supported_boot_mode(
@@ -726,6 +1018,15 @@ class RedfishOperations(operations.IloOperations):
                      'logical_nvdimm_n': (
                      json.dumps(memory_data.has_logical_nvdimm_n))})
 
+            gpu_capabilities = gpu_common.gpu_capabilities(sushy_system,
+                                                           sushy_chassis)
+            for member in gpu_capabilities:
+                for key in member:
+                    capabilities.update(member.get(key))
+
+            capabilities.update(
+                self._parse_security_dashboard_values_for_capabilities())
+
         except sushy.exceptions.SushyError as e:
             msg = (self._("The Redfish controller is unable to get "
                           "resource or its members. Error "
@@ -1115,6 +1416,19 @@ class RedfishOperations(operations.IloOperations):
         sushy_system = self._get_sushy_system(PROLIANT_SYSTEM_ID)
         return sushy_system.has_disk_erase_completed()
 
+    def do_one_button_secure_erase(self):
+        """Perform the one button secure erase on the hardware.
+
+        The One-button secure erase process resets iLO and deletes all licenses
+        stored there, resets BIOS settings, and deletes all AHS and warranty
+        data stored on the system. It also erases supported non-volatile
+        storage data and deletes any deployment settings profiles.
+
+        :raises: IloError, on an error from iLO.
+        """
+        sushy_system = self._get_sushy_system(PROLIANT_SYSTEM_ID)
+        sushy_system.do_one_button_secure_erase()
+
     def get_current_bios_settings(self, only_allowed_settings=False):
         """Get current BIOS settings.
 
@@ -1339,17 +1653,33 @@ class RedfishOperations(operations.IloOperations):
                  not supported on the server.
         """
         sushy_system = self._get_sushy_system(PROLIANT_SYSTEM_ID)
+
         if(self._is_boot_mode_uefi()):
             cert_list = []
             for cert_file in cert_file_list:
                 with open(cert_file, 'r') as f:
                     data = json.dumps(f.read())
                 p = re.sub(r"\"", "", data)
-                q = re.sub(r"\\n", "\r\n", p)
-                r = q.rstrip()
-                cert = {}
-                cert['X509Certificate'] = r
-                cert_list.append(cert)
+                q = re.sub(r"\\n", "\r\n", p).rstrip()
+
+                c_list = re.findall(_CERTIFICATE_PATTERN, q, re.DOTALL)
+
+                if len(c_list) == 0:
+                    LOG.warning("Could not find any valid certificate in "
+                                "%(cert_file)s. Ignoring." %
+                                {"cert_file": cert_file})
+                    continue
+
+                for content in c_list:
+                    cert = {}
+                    cert['X509Certificate'] = content
+                    cert_list.append(cert)
+
+            if len(cert_list) == 0:
+                msg = (self._("No valid certificate in %(cert_file_list)s.") %
+                       {"cert_file_list": cert_file_list})
+                LOG.debug(msg)
+                raise exception.IloError(msg)
 
             cert_dict = {}
             cert_dict['NewCertificates'] = cert_list
@@ -1366,34 +1696,79 @@ class RedfishOperations(operations.IloOperations):
             msg = 'TLS certificate cannot be upload in BIOS boot mode'
             raise exception.IloCommandNotSupportedInBiosError(msg)
 
-    def remove_tls_certificate(self, fp_list):
+    def remove_tls_certificate(self, cert_file_list=[]):
         """Removes the TLS certificate from the iLO.
 
-        :param fp_list: List of finger prints of the TLS certificates
+        :param cert_file_list: List of TLS certificate files
 
         :raises: IloError, on an error from iLO.
         :raises: IloCommandNotSupportedError, if the command is
                  not supported on the server.
         """
         sushy_system = self._get_sushy_system(PROLIANT_SYSTEM_ID)
-        if(self._is_boot_mode_uefi()):
-            cert = {}
-            del_cert_list = []
-            for fp in fp_list:
+
+        if not self._is_boot_mode_uefi():
+            msg = 'TLS certificates cannot be removed in BIOS boot mode'
+            raise exception.IloCommandNotSupportedInBiosError(msg)
+
+        cert_dict = {}
+        del_cert_list = []
+
+        if not cert_file_list:
+            tls_certificates = (sushy_system.bios_settings.tls_config.
+                                tls_certificates)
+            for cert in tls_certificates:
+                fp = cert.get("FingerPrint")
                 cert_fp = {
                     "FingerPrint": fp
                 }
                 del_cert_list.append(cert_fp)
-            cert.update({"DeleteCertificates": del_cert_list})
-            try:
-                (sushy_system.bios_settings.tls_config.
-                 tls_config_settings.remove_tls_certificate(cert))
-            except sushy.exceptions.SushyError as e:
-                msg = (self._("The Redfish controller has failed to remove "
-                              "TLS certificate. Error %(error)s") %
-                       {'error': str(e)})
-                LOG.debug(msg)
-                raise exception.IloError(msg)
+
         else:
-            msg = 'TLS certificate cannot be removed in BIOS boot mode'
-            raise exception.IloCommandNotSupportedInBiosError(msg)
+            for cert_file in cert_file_list:
+                with open(cert_file, 'r') as f:
+                    data = json.dumps(f.read())
+                    p = re.sub(r"\"", "", data)
+                    q = re.sub(r"\\n", "\r\n", p).rstrip()
+
+                    c_list = re.findall(_CERTIFICATE_PATTERN, q, re.DOTALL)
+
+                    if len(c_list) == 0:
+                        LOG.warning("Could not find any valid certificate in "
+                                    "%(cert_file)s. Ignoring." %
+                                    {"cert_file": cert_file})
+                        continue
+
+                    for content in c_list:
+                        pem_lines = [line.strip() for line in (
+                            content.strip().split('\n'))]
+
+                        try:
+                            der_data = b64decode(''.join(pem_lines[1:-1]))
+                        except ValueError:
+                            LOG.warning("Illegal base64 encountered "
+                                        "in the certificate.")
+                        else:
+                            cert = load_certificate(FILETYPE_ASN1, der_data)
+                            fp = cert.digest('sha256').decode('ascii')
+                            cert_fp = {
+                                "FingerPrint": fp
+                            }
+                            del_cert_list.append(cert_fp)
+
+        if len(del_cert_list) == 0:
+            msg = (self._("No valid certificate in %(cert_file_list)s.") %
+                   {"cert_file_list": cert_file_list})
+            raise exception.IloError(msg)
+
+        cert_dict.update({"DeleteCertificates": del_cert_list})
+
+        try:
+            (sushy_system.bios_settings.tls_config.
+             tls_config_settings.remove_tls_certificate(cert_dict))
+        except sushy.exceptions.SushyError as e:
+            msg = (self._("The Redfish controller has failed to remove "
+                          "TLS certificate. Error %(error)s") %
+                   {'error': str(e)})
+            LOG.debug(msg)
+            raise exception.IloError(msg)
